=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2014 Fredo6 - Designed and written Mar 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6OriginTargetPicker.rb
# Original Date	:   11 Mar 2014 
# Description	:   Manage the picking of an origin and a target with second-level inferences
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#-------------------------------------------
# Text for tooltips of Inferences
#-------------------------------------------

T6[:INFER_InGroup] = "in Group"
T6[:INFER_InComponent] = "in Component"
T6[:INFER_InImage] = "in Image"
T6[:INFER_Cpoint] = "Guide Point"
T6[:INFER_Cline] = "On Guide Line"
T6[:INFER_Endpoint] = "Endpoint"
T6[:INFER_Midpoint] = "Midpoint"
T6[:INFER_CenterFace] = "Center of Face"
T6[:INFER_CenterArc] = "Center of arc"
T6[:INFER_CenterComp] = "Center of Component"
T6[:INFER_CenterCompTop] = "Center of Top Component"
T6[:INFER_CenterGroup] = "Center of Group"
T6[:INFER_CenterGroupTop] = "Center of Top Group"
T6[:INFER_CenterPolygon] = "Center of Circle/Polygon"
T6[:INFER_OnEdge] = "on Edge"
T6[:INFER_OnFace] = "on Face"
T6[:INFER_AtOrigin] = "at Origin"
T6[:INFER_OnXAxis] = "on X axis"
T6[:INFER_OnYAxis] = "on Y axis"
T6[:INFER_OnZAxis] = "on Z axis"
T6[:INFER_AlongXAxis] = "Along X axis"
T6[:INFER_AlongYAxis] = "Along Y axis"
T6[:INFER_AlongZAxis] = "Along Z axis"
T6[:INFER_AlongXAxisComp] = "Along X axis of Component"
T6[:INFER_AlongYAxisComp] = "Along Y axis of Component"
T6[:INFER_AlongZAxisComp] = "Along Z axis of Component"
T6[:INFER_PerpEdgeFace] = "Perpendicular to Edge and Face"
T6[:INFER_AlongEdgeDirection] = "Along Edge direction"
T6[:INFER_PerpTwoEdges] = "Perpendicular to two Edges"
T6[:INFER_AlongCline] = "Along Guide line"
T6[:INFER_AlongAxesBbox] = "Along axes of Bounding Box"
T6[:INFER_AlongNormalFace] = "Along Normal to Face"
T6[:INFER_InterEdgeFace] = "Intersection Edge - Face"
T6[:INFER_InterClineFace] = "Intersection Guide Line - Face"
T6[:INFER_InterClineEdge] = "Intersection Guide Line - Edge"
T6[:INFER_InterClineCline] = "Intersection Guide Line - Guide Line"
T6[:INFER_InterVectorFace] = "Intersection Direction and Face"
T6[:INFER_InterPlaneEdge] = "Intersection Plane and Edge"
T6[:INFER_TextAnchor] = "Text Anchor"
T6[:INFER_OnTextLabel] = "on Text Label"
T6[:INFER_OnTextLeader] = "on Text Leader"
T6[:INFER_OnDimension] = "on Dimension"
T6[:INFER_OnDrawingElement] = "on Drawing Element"
T6[:INFER_AtCornerBbox] = "at Corner of Bounding Box"
T6[:INFER_OnFaceBbox] = "on Face of Bounding Box"

#=============================================================================================
#=============================================================================================
# Class MoveAlongTool: main class for the Interactive tool
#=============================================================================================
#=============================================================================================

class OriginTargetPicker

#INIT: Class instance initialization
def initialize(*hargs)
	#Initialization
	@tr_id = Geom::Transformation.new
	@model = Sketchup.active_model
	@selection = @model.selection
	@rendering_options = Sketchup.active_model.rendering_options
	@view = @model.active_view
	@ph = @view.pick_helper	
	@ip = Sketchup::InputPoint.new
	@ogl = Traductor::OpenGL_6.new	
	
	#Initialization
	init_colors
	
	#Parsing the arguments
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash
	end
	
	#Precision parameters
	@precision_default = 8
	@inference_angle_default = 8.degrees
	@precision = @precision_default unless @precision
	@inference_angle = @inference_angle_default unless @inference_angle
	@inference_distance = [3 * @precision, 25].min
	@inference_distance = 15 if @inference_distance < 15
	
	@ip_origin = Sketchup::InputPoint.new unless @ip_origin
	@ip_target = Sketchup::InputPoint.new unless @ip_target
	
	#Initialization
	mark_init
end

#INIT: Parse the parameters of the Face Picker
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /precision/i
		@precision = value
	when /inference_angle/i
		@inference_angle = value.degrees
	when /color_box_distance/
		@color_box_distance = value
	when /ignore_selection/i
		@ignore_selection = value
	end	
end	

#INIT: Set the precision and inference angle
def set_precision(precision) ; @precision = (precision) ? precision : @precision_default ; end
def set_inference_angle(angle) ; @inference_angle = (angle) ? angle.degrees : @inference_angle_default ; end

#INIT: reset the environment
def reset
	@pt_origin = @origin_info = nil
	@pt_target = @target_info = nil
	@inference_info = nil
	@hi_curve_pts = nil
	@top_parent = @info_parent = nil
	inference_reset_centers
end

#INIT: Initialize texts and messages
def init_colors
	lst_color_dir = [ [:free, 'black'], [:lock, 'black'], [:follow_bbox, 'slategray'],
	                  [:follow_x, 'red'], [:follow_y, 'green'], [:follow_z, 'blue'], [:perp_edge, 'brown'],
					  [:follow_edge, 'magenta'], [:follow_cline, 'purple'], [:follow_normal, 'darkorange'],
					  [:follow_x_comp, 'tomato'], [:follow_y_comp, 'limegreen'], [:follow_z_comp, 'skyblue'],
					  [:plane_of_edges, 'goldenrod']]
	@color_box_distance = ['lightgrey', 'black']

	#Color for vectors and planes
	@hsh_color_vector = Hash.new 'black'
	@hsh_bkcolor_plane = Hash.new 'gray'
	@hsh_frcolor_plane = Hash.new 'black'
	lst_color_dir.each do |symb, color|
		@hsh_color_vector[symb] = color
		pcolor = Sketchup::Color.new color
		pcolor.alpha = 0.1
		@hsh_bkcolor_plane[symb] = pcolor
		@hsh_frcolor_plane[symb] = color
	end
end

#INIT: Reset the centers for inference
def rollback
	inference_reset_centers
end

#---------------------------------------------------------
# PRESELECTION: Manage preselection
#---------------------------------------------------------
	
#PRESELECTION: set a preselection for the picker tool	
def preselection_set(preselection)
	@active_selection = @preselection = preselection
	@cumulative_selection = nil
	suselection_add(@preselection, true)
	@preselection
end

#PRESELECTION: Get the preselection (nil if none)
def preselection_get
	@preselection || @cumulative_selection
end

#PRESELECTION: Restore the preselection if any
def preselection_restore
	@preselection = @cumulative_selection unless @preselection	
	return nil unless @preselection
	preselection = @preselection.find_all { |e| e.valid? }
	preselection_set preselection
end
	
#PRESELECTION: Cumulate the selection
def preselection_cumulate
	@cumulative_selection = @selection.to_a
end

#PRESELECTION: Set the preselection when there is a cumulative selection
def preselection_set_when_cumulative
	if @cumulative_selection
		suselection_add @cumulative_selection, true
		preselection_cumulate
	end	
end

#---------------------------------------------------------
# MARK: Manage marks
#---------------------------------------------------------

#MARK: Initialize the instructions for marks (at origin)
def mark_init
	@hsh_marks = {}
	@color_comp = 'mediumorchid'
	@marks_keep_color = [:cpoint, :on_cline, :on_text, :text_leader, :text_anchor, :center_face, 
	                     :center_comp_mark, :center_comp_mark_top, :center_comp, :center_comp_top, :center_group, :center_group_top,
						 :bbox_corner, :bbox_face, :inter_plane_edge, :inter_vector_face]
	@bbox_lines_indexing = [0, 1, 1, 3, 3, 2, 2, 0, 0, 4, 1, 5, 2, 6, 3, 7, 4, 5, 5, 7, 6, 7, 4, 6]
	
	#Default mark
	dec = 3
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[nil] = [[GL_POLYGON, pts, 'darkred']]

	#Mark for void origin
	dec = 3
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:void_origin] = [[GL_POLYGON, pts, 'green']]

	#Mark on Face
	dec = 6
	pts = [[dec, 0], [0, dec], [-dec, 0], [0, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:on_face] = [[GL_POLYGON, pts, 'blue'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark on Edge
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:on_edge] = [[GL_POLYGON, pts, 'red'], [GL_LINE_LOOP, pts, 'white', 1, '']]
	
	#Mark on Text
	dec = 6
	pts = [[dec, 0], [0, dec], [-dec, 0], [0, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:on_text] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'yellow', 1, '']]

	#Mark on Text
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:text_leader] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'yellow', 1, '']]
	
	#Mark on Guide line
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:on_cline] = [[GL_POLYGON, pts, 'darkgray'], [GL_LINE_LOOP, pts, 'white', 1, '']]
	
	#Mark at Midpoint
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:midpoint] = [[GL_POLYGON, pts, 'red'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'white', 1, '']]

	#Mark at Center of Face
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:center_face] = [[GL_POLYGON, pts, 'blue'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'white', 1, '']]

	#Mark at Vertex
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:vertex] = [[GL_POLYGON, pts, 'green'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark at Guide Point
	dec = 5
	dec1 = dec - 1
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec1, 0], [dec1, 0], [0, -dec1], [0, dec1]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:cpoint] = [[GL_POLYGON, pts, 'darkgray'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'black', 2, '']]

	#Mark at Text Anchor
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:text_anchor] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark at origin
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:origin] = [[GL_POLYGON, pts, 'yellow'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark on axes
	dec = 6
	pts = [[-dec, -dec], [dec, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	[[:x, 'red'], [:y, 'green'], [:z, 'blue']].each do |symb, color|
		@hsh_marks[symb] = [[GL_POLYGON, pts, color], [GL_LINE_LOOP, pts, 'white', 1, '']]
	end	

	#Mark at Intersection
	dec = 5
	dec1 = dec-1
	pts = [[-dec, -dec], [dec, dec], [-dec, dec], [dec, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	pts2 = [[-dec1, -dec1], [dec1, dec1], [-dec1, dec1], [dec1, -dec1]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	inter_mark = @hsh_marks[:inter] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'red', 2, '']]
	[:inter_edge_face, :inter_cline_face, :inter_cline_edge, :inter_cline_cline].each do |symb|
		@hsh_marks[symb] = inter_mark
	end	
	@hsh_marks[:inter_vector_face] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'navy', 2, '']]
	@hsh_marks[:inter_plane_edge] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'tomato', 2, '']]
	
	#Mark at Center of Arc curve
	dec = 5
	pts = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:center_arc_mark] = [[GL_LINES, pts, 'red', 1, '']]
	pts2 = G6.pts_circle 0, 0, dec, 6
	@hsh_marks[:center_arc] = @hsh_marks[:center_polygon] = [[GL_POLYGON, pts2, 'red'], [GL_LINES, pts, 'white', 1, '']]

	#Mark at Center of Component or Group
	dec = 6
	pts1 = G6.pts_circle 0, 0, dec
	pts2 = G6.pts_circle 0, 0, dec-2
	pts3 = G6.pts_circle 0, 0, dec-4
	[[:center_comp_mark_top, ['red', 'yellow', 'blue']], [:center_comp_top, ['blue', 'yellow', 'red']],
	 [:center_comp, ['black', 'white', 'orange']], [:center_comp_mark, ['orange', 'white', 'black']],
	 [:bbox_corner, ['blue', 'white', 'blue']]].each do |a|
		symb, colors = a
		@hsh_marks[symb] = [[GL_POLYGON, pts1, colors[0]], [GL_POLYGON, pts2, colors[1]], [GL_POLYGON, pts3, colors[2]]]
	end	
	@hsh_marks[:center_group] = @hsh_marks[:center_comp]
	@hsh_marks[:center_group_top] = @hsh_marks[:center_comp_top]
end

#MARK: Compute the instructions 2D for drawing at point pt3d
def mark_draw(view, info)
	pt3d, mark, info_comp = info
	return unless pt3d
	
	#Transformation to 2d
	pt2d = view.screen_coords pt3d
	t = Geom::Transformation.translation pt2d
	
	#Original instructions
	#mark = nil unless @hsh_marks.has_key?(mark)
	return unless @hsh_marks.has_key?(mark)
	ls = @hsh_marks[mark]
	
	#Creating the instructions for drawing
	in_comp = (info_comp && info_comp[1])
	instructions = []
	ls.each do |a|
		code, pts, color, width, stipple = a
		color = @color_comp if in_comp && code == GL_POLYGON && mark && !@marks_keep_color.include?(mark)
		instructions.push [code, pts, color, width, stipple]
	end	
	@ogl.process_draw_GL view, t, instructions
end

#MARK: Get the current tooltip based on inference or direction
def get_view_tooltip
	@view_tooltip
end
	
#---------------------------------------------------------
# EXCLUSION: Manage the elements to be excluded
#---------------------------------------------------------

#EXCLUSION: Register the exclusion
def exclusion_register(lst_elts=nil)
	@lst_exclusion = (lst_elts.class == Array) ? lst_elts.clone : nil
	
	@hsh_exclusion_ids = {}
	@hsh_snapshot_ids = {}
	@hsh_compdef = {}
	@at_top_level = false
	return unless @lst_exclusion
	@lst_exclusion = @lst_exclusion.find_all { |e| e.valid? }
	exclusion_map_entityID(@lst_exclusion)	
	
	#Registering ids at top level
	ee = @model.active_entities
	(ee.grep(Sketchup::Face) + ee.grep(Sketchup::Edge) + ee.grep(Sketchup::ConstructionPoint)).each do |e|
		@hsh_snapshot_ids[e.entityID] = e
	end	
end

#EXCLUSION: Go through the selection and map the entity id		
def exclusion_map_entityID(lentities=nil)
	lentities.each do |e|
		next unless e.valid?
		@hsh_exclusion_ids[e.entityID] = e
		if e.instance_of?(Sketchup::ComponentInstance)
			lst = e.definition.entities.to_a
		elsif e.instance_of?(Sketchup::Group)
			lst = e.entities.to_a
		else 		
			(e.edges + e.faces).each { |ee| @hsh_exclusion_ids[ee.entityID] = ee } if e.instance_of?(Sketchup::Vertex)
			e.faces.each { |ee| @hsh_exclusion_ids[ee.entityID] = ee } if e.instance_of?(Sketchup::Edge)
			@at_top_level = true
			next
		end
		lst = lst.grep(Sketchup::ComponentInstance) + lst.grep(Sketchup::Group) + lst.grep(Sketchup::Image)
		exclusion_map_entityID(lst) unless lst.empty?
	end	
end

#EXCLUSION: Exclude when picked element is part of the preselection
def part_of_exclusion?(lentities, parent)
	#Element is part of a grouponent
	return true if parent && @hsh_exclusion_ids[parent.entityID]
	
	#Element is at toplevel model
	lentities.each do |e|
		next unless e
		id = e.entityID
		return true if @hsh_exclusion_ids[id] || (!parent && !@hsh_snapshot_ids[id])
		if e.instance_of?(Sketchup::Vertex)
			return true if e.edges.find { |edge| @hsh_exclusion_ids[edge.entityID] }
			return true if e.faces.find { |face| @hsh_exclusion_ids[face.entityID] }
		end
		if e.instance_of?(Sketchup::Face)
			return true if e.vertices.find { |vx| @hsh_exclusion_ids[vx.entityID] }
		end
		if e.instance_of?(Sketchup::Edge)
			return true if [e.start, e.end].find { |vx| @hsh_exclusion_ids[vx.entityID] }
		end
		return true if (e.instance_of?(Sketchup::Edge) || e.instance_of?(Sketchup::Face)) && e.vertices.find { |vx| @lst_exclusion.include?(vx) }
	end
	false
end

#---------------------------------------------------------
# DIRECTION: Manage forced_direction
#---------------------------------------------------------

#DIRECTION: Set the imposed direction as a vector
def set_vector_direction(vec, code=nil)
	@forced_vector_code = code if code
	@forced_plane_dir = nil
	@forced_vector_dir = vec
	@direction_tooltip = anchor_tooltip([nil, code])
end

#DIRECTION: Set the imposed direction as a vector
def set_plane_direction(vec, code=nil)
	@forced_plane_code = code
	@forced_vector_dir = nil
	@forced_plane_dir = vec
	@direction_tooltip = anchor_tooltip([nil, code])
end

#DIRECTION: return the tooltip associated to the direction or nil if none
def direction_tooltip
	@direction_tooltip
end

#DIRECTION: Compute the target when direction is forced
def direction_adjust_target(view, x, y)
	@pt_target, type, info_comp = @target_info
	@info_drive3d = @target_info
	@pt_drive2d = Geom::Point3d.new x, y, 0
	return if @pt_origin == @pt_target
	
	#Vector direction imposed
	if @forced_vector_dir
		#Position locked on a point, edge, or surface in the model
		if @pt_target
			#Force intersection with face
			if type == :on_face
				face, comp, tr = info_comp
				plane = [tr * face.vertices[0].position, G6.transform_vector(face.normal, tr)]
				ptinter = Geom.intersect_line_plane [@pt_origin, @forced_vector_dir], plane
				if ptinter && [1, 2, 4].include?(face.classify_point(tr.inverse * ptinter))
					@pt_target = ptinter
					type = :inter_vector_face
					@target_info = [@pt_target, type, [face, comp, tr]]
					return
				else
					@pt_target = @pt_target.project_to_line [@pt_origin, @forced_vector_dir]
				end	
			else
				@pt_target = @pt_target.project_to_line [@pt_origin, @forced_vector_dir]
			end	

		#Position is free
		else
			pvorig = view.screen_coords @pt_origin
			pv0 = view.screen_coords @pt_origin.offset(@forced_vector_dir, 100)
			ptxy = Geom::Point3d.new x, y, 0
			pv1 = ptxy.project_to_line [pvorig, pv0]
			a = Geom.closest_points [@pt_origin, @forced_vector_dir], view.pickray(pv1.x, pv1.y)
			@pt_target = a[0]
		end	
		
	#Planar direction imposed	
	elsif @forced_plane_dir
		
		#No target picked
		unless @pt_target
			pt = Geom.intersect_line_plane view.pickray(x, y), [@pt_origin, @forced_plane_dir]
			@pt_target = (pt) ? pt : @pt_origin
			@target_info = @pt_target, type, info_comp		
		end
		
		#Imposed plane
		plane = [@pt_origin, @forced_plane_dir]
		
		#No good inference for intersection
		unless [:on_edge, :on_cline, :x, :y, :z, :midpoint, :vertex].include?(type)
			@pt_target = @pt_target.project_to_plane plane 
			@target_info = @pt_target, type, info_comp		
			return
		end	
		
		#Forcing intersection of plane with Edge (defined by edge, midpoint or vertex
		if type == :on_edge || type == :midpoint || type == :vertex
			if type == :vertex
				vx, comp, tr = info_comp
				ledges = vx.edges
				if ledges.length > 2
					@pt_target = @pt_target.project_to_plane plane
					@target_info = @pt_target, type, info_comp
					return
				end	
				edge = ledges.first
			else
				edge, comp, tr = info_comp
			end	
			new_target, edge1 = direction_intersect_edge_plane(edge, tr, plane)
			if new_target
				@pt_target = new_target
				type = :inter_plane_edge
				@target_info = [@pt_target, type, [edge1, comp, tr]]
				return
			elsif type == :vertex
				@pt_target = @pt_target.project_to_plane plane
				@target_info = @pt_target, type, info_comp
				return
			else
				line = [tr * edge.start.position, tr * edge.end.position]
			end	

		#Forcing intersection of plane with Guide line
		elsif type == :on_cline
			cline, comp, tr = info_comp
			line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]

		#Forcing intersection of axes
		elsif type == :x
			line = [@pt_origin, X_AXIS]
		elsif type == :y
			line = [@pt_origin, Y_AXIS]
		elsif type == :z
			line = [@pt_origin, Z_AXIS]
		end	
		
		ptinter = Geom.intersect_line_plane line, plane
		@pt_target = (ptinter) ? ptinter : @pt_target.project_to_plane(plane)
		
	#Handling Privileged directions
	elsif direction_privileged_on_target?(view, x, y)
		return
	end
	
	#Configuring the new target information
	@target_info = @pt_target, type, info_comp
end

#DIRECTION:
def direction_privileged_on_target?(view, x, y)
	@pt_target, type, info_comp = @target_info
	pixels = 2 * @precision
	
	#Parameters for the privileged direction
	info_dir = anchor_follow_direction(view, x, y, @pt_origin)
	ptdir, type_dir = info_dir
	
	#Not a privileged direction
	return false unless type_dir && type_dir.to_s =~ /follow/
	
	#Exploring the potential intersections
	
	#With a face
	if type == :on_face
		line_dir = [@pt_origin, ptdir]
		face, comp, tr = info_comp
		plane = [@pt_target, G6.transform_vector(face.normal, tr)]
		ptinter = Geom.intersect_line_plane line_dir, plane
		if ptinter && [1, 2, 4].include?(face.classify_point(tr.inverse * ptinter)) && 
		   G6.points_close_in_pixel?(view, ptinter, @pt_target, pixels)
			@target_info = [ptinter, type, info_comp]
			@info_drive3d = [ptinter, type_dir]
			return true
		end
		@info_drive3d = @target_info = info_dir
		return true
	end
	
	#With an edge	
	if type == :on_edge
		edge, comp, tr = info_comp
		line = [tr * edge.start.position, tr * edge.end.position]
		
	elsif type == :x
		line = [ORIGIN, X_AXIS]
	elsif type == :y
		line = [ORIGIN, Y_AXIS]
	elsif type == :z
		line = [ORIGIN, Z_AXIS]
	elsif type == :on_cline
		cline, comp, tr = info_comp
		line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
	else
		line = nil
	end
	
	if line
		line_dir = [@pt_origin, ptdir]
		ptinter = Geom.intersect_line_line line, line_dir
		if ptinter && G6.points_close_in_pixel?(view, ptinter, @pt_target, pixels)
			@target_info = [ptinter, type, info_comp]
			@info_drive3d = [ptinter, type_dir]
			return true	
		end
	end
	
	if type == :free
		@info_drive3d = @target_info = info_dir
		return true
	end
	false
end

#DIRECTION: Find the intersection of an edge and neighbours with the forced plane
def direction_intersect_edge_plane(edge, tr, plane)
	pt1 = tr * edge.start.position
	pt2 = tr * edge.end.position
	ptinter = Geom.intersect_line_plane [pt1, pt2], plane
	
	#Intersection found
	if ptinter && G6.point_within_segment?(ptinter, pt1, pt2)
		return [ptinter, edge]
	end	
		
	#Getting edge on the right
	edge_right = edge_left = edge
	
	for i in 0..5
		ptinter = nil
		ptinter, edge_right = direction_intersect_edge_plane_next(edge_right, tr, plane, edge_right.end) if edge_right
		return [ptinter, edge_right] if ptinter
		ptinter, edge_left = direction_intersect_edge_plane_next(edge_left, tr, plane, edge_left.start) if edge_left
		return [ptinter, edge_left] if ptinter
	end
	
	#Prolonging the Edge if part of an alone sequence
	nil	
end

#DIRECTION: Find the intersection of the next edge the forced plane
def direction_intersect_edge_plane_next(edge, tr, plane, vx)
	ledges = vx.edges.find_all { |e| e != edge }
	len = ledges.length
	ledges = ledges.find_all { |e| G6.edge_plain?(e) } if len > 1
	return nil if ledges.length != 1
	edge1 = ledges.first

	pt1 = tr * edge1.start.position
	pt2 = tr * edge1.end.position
	ptinter = Geom.intersect_line_plane [pt1, pt2], plane
	
	#Intersection found
	(ptinter && G6.point_within_segment?(ptinter, pt1, pt2)) ? [ptinter, edge1] : [nil, edge1]
end

#DIRECTION: Compute the symbol of direction based on vector and specs
def direction_symbol(vec, code=nil)
	return 'black' unless vec
	unless code
		return @forced_vector_code if @forced_vector_dir
		return @forced_plane_code if @forced_plane_dir
	end
	
	if vec.parallel?(X_AXIS)
		symb = :x
	elsif vec.parallel?(Y_AXIS)
		symb = :y
	elsif vec.parallel?(Z_AXIS)
		symb = :z
	elsif code
		symb = code
	else	
		symb = :default
	end
	symb
end
	
#DIRECTION: Compute privileged direction from a point of reference	
def compute_privileged_directions(reference_info)
	ptref, type_ref, info_comp_ref = reference_info
	
	#Axes
	ls_priv = [[:follow_z, Z_AXIS], [:follow_x, X_AXIS], [:follow_y, Y_AXIS]]
	
	#Axes of components if different from model axis
	@top_axes = false
	elt, comp, tr = info_comp_ref
	if @top_parent.instance_of?(Sketchup::ComponentInstance)
		tr = @top_parent.transformation
		[[:follow_z_comp, Z_AXIS], [:follow_x_comp, X_AXIS], [:follow_y_comp, Y_AXIS]].each do |symb, vec|
			vec_comp = tr * vec
			unless vec_comp.parallel?(vec)
				ls_priv.push [symb, vec_comp]
				@top_axes = true
			end
		end
	end
	
	#Follow a privileged direction
	symb = nil
	case type_ref
	when :on_edge, :midpoint
		edge, comp, tr = info_comp_ref
		vec_edge = (tr * edge.start.position).vector_to(tr * edge.end.position)
		ls_priv.push [:follow_edge, vec_edge]
		edge.faces.each do |face|
			normal = G6.transform_vector(face.normal, tr)
			vec = vec_edge * normal
			ls_priv.push [:perp_edge, vec]
		end
	
	when :on_face, :center_face
		face, comp, tr = info_comp_ref
		ls_priv.push [:follow_normal, G6.transform_vector(face.normal, tr)]
	
	when :vertex
		vx, comp, tr = info_comp_ref
		edges = vx.edges
		edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
		if edges.length == 1
			edge = edges[0]
			ls_priv.push [:follow_edge, (tr * edge.start.position).vector_to(tr * edge.end.position)]
		elsif edges_plain.length == 1	
			edge = edges_plain[0]
			ls_priv.push [:follow_edge, (tr * edge.start.position).vector_to(tr * edge.end.position)]
		else edges_plain.length == 2
			edges_plain.each do |edge|
				vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
				ls_priv.push [:follow_edge, vec] if vec.valid?
			end	
		end
		
	when :on_cline
		cline, comp, tr = info_comp_ref
		ls_priv.push [:follow_cline, G6.transform_vector(cline.direction, tr)]
		
	when :bbox_corner, :center_comp, :center_group
		comp, parent, tr = info_comp_ref
		[X_AXIS, Y_AXIS, Z_AXIS].each do |vec|
			ls_priv.push [:follow_bbox, G6.transform_vector(vec, tr)]
		end	
	end
	
	@origin_privileged_directions = ls_priv
end
	
#DIRECTION: Compute a vector direction from picking elements in the model	
def direction_vector_from_model(info)
	pt, symb, info_elt = info
	elt, comp, tr = info_elt
	vec = nil
	
	case symb
	when :on_edge, :midpoint
		vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
		code = :follow_edge
	when :on_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal	
	end
	
	(vec) ? [vec, code] : nil
end

#DIRECTION: Compute a vector direction from picking elements in the model	
def direction_plane_from_model(info)
	pt, symb, info_elt = info
	elt, comp, tr = info_elt
	vec = nil
	
	case symb
	when :on_edge, :midpoint
		vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
		code = :follow_edge
	when :on_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal		
	when :vertex
		vx = elt
		edges = vx.edges
		edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
		if edges.length == 1
			edge = edges[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
		elsif edges_plain.length == 1	
			edge = edges_plain[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
		else edges_plain.length == 2
			edge1, edge2 = edges_plain
			vec1 = (tr * edge1.start.position).vector_to(tr * edge1.end.position)
			vec2 = (tr * edge2.start.position).vector_to(tr * edge2.end.position)
			vec = vec1 * vec2
			code = :plane_of_edges			
		end
		
	end
	
	(vec && vec.valid?) ? [vec, code] : nil
end
	
#---------------------------------------------------------
# ORIGIN: Manage the origin
#---------------------------------------------------------

#ORIGIN: Pick the target (complex mode with exclusion)
def origin_pick(view, x, y, extended, multi_selection, no_vertex)
	@mode_origin = true
	@origin_privileged_directions = nil
	@inference_info = @origin_info = anchor_pick(view, x, y)
	if @origin_info
		@pt_origin = @origin_info.first
		compute_privileged_directions(@origin_info)
	else
		@ip.pick view, x, y
		pt = Geom.intersect_line_plane view.pickray(x, y), [@ip.position, view.camera.direction]
		@origin_info = [pt, :void_origin, nil]
		@pt_origin = nil
	end
	
	#Auto selection
	@active_selection = auto_selection(view, x, y, extended, multi_selection, no_vertex)

	#Setting the tooltip for the view
	@view_tooltip = anchor_tooltip(@origin_info)
	view.tooltip = (@view_tooltip) ? '   ' + @view_tooltip : nil

	@pt_origin
end
	
#ORIGIN: Active selection when picking origin	
def origin_active_selection
	@active_selection
end
	
#ORIGIN: get information on origin	
def origin_info
	@origin_info
end
	
#ORIGIN: get information on the top parent picked in Origin mode	
def origin_info_top_parent
	@info_top_parent
end
		
#---------------------------------------------------------
# TARGET: Manage the target
#---------------------------------------------------------
	
#TARGET: Pick the target (complex mode with exclusion)
def pick_target(view, x, y, with_exclusion=false)
	@mode_origin = false
	
	#Setting the reference at origin
	@inference_info = @origin_info
	
	#Picking the target
	@target_info = anchor_pick(view, x, y, with_exclusion)
	@pt_target, type, info_comp = @target_info
	
	#Adjusting the target based on a possible forced direction or plane
	direction_adjust_target view, x, y
			
	#Setting the tooltip for the view
	@view_tooltip = anchor_tooltip(@target_info)
	view.tooltip = (@view_tooltip) ? '   ' + @view_tooltip : nil

	@pt_target, type, info_comp = @target_info
	@pt_drive3d, = @info_drive3d

	@pt_target
end

#TARGET: get information on target	
def target_info
	@target_info
end
	
#---------------------------------------------------------
# ANCHOR: Manage the anchor point detection
#---------------------------------------------------------

#ANCHOR: Top level method to find an anchor
def anchor_pick(view, x, y, with_exclusion=false)
	#Picking the point
	@ph.do_pick x, y, 8
	
	#Finding an element which is not part of the selection
	parent = nil
	lst_anchor = []
	for i in 0..@ph.count
		path = @ph.path_at(i)
		break unless path && path.length > 0
		elt = path.last
		
		comp = (path.length == 1) ? nil : path[-2]
		unless with_exclusion && part_of_exclusion?([elt], comp)
			lst_anchor.push [i, elt, comp, @ph.transformation_at(i)]
		end	
	end
	
	#Analysing the anchors
	info = nil
	top_parent = nil
	unless lst_anchor.empty?
		info = anchor_analysis(view, lst_anchor, x, y)
		i = lst_anchor[0][0]
		top_parent = @ph.path_at(i)[0]
		top_parent = nil unless top_parent.instance_of?(Sketchup::ComponentInstance) || top_parent.instance_of?(Sketchup::Group) ||
		                        top_parent.instance_of?(Sketchup::Image)
	end
	
	#Anchor not found or on Text or Dimension (if so, privilege directions)
	info_def = nil
	if info
		pt, symb = info
		info_def = info if symb.to_s =~ /text/ || symb == :dimension || symb == :element
	end	
	
	#Calculating other inferences and taking the closest to camera
	info_void = anchor_in_void(view, x, y, info_def)
	info = anchor_compare_with_void(view, info, info_void)
	info = info_def unless info
	
	#Computing inferences for Center of active face or arc, component or group
	target, symb, info_comp = info
	if info && symb.to_s !~ /center/
		elt, comp, tr = info_comp
		if elt.instance_of?(Sketchup::Face)
			inference_center_face(elt, comp, tr)
		elsif elt.instance_of?(Sketchup::Edge) && 
			inference_center_arc_from_edge(elt, comp, tr)
		elsif elt.instance_of?(Sketchup::Vertex)
			inference_center_arc_from_vertex(elt, comp, tr)
		end
		
		#Computing the centers in the parent group or component
		if ![:center_comp, :center_comp_top, :center_group, :center_group_top, :bbox_corner, :bbox_face_center].include?(info[1])
			if top_parent != comp
				if comp
					inference_center_grouponent(top_parent, comp, tr)	
					inference_bbox_grouponent(top_parent, comp, tr)
				end	
			elsif comp
				inference_center_grouponent(comp, nil, @tr_id)	
				inference_bbox_grouponent(comp, nil, @tr_id)	
			end
		end			
	end
	
	#Computing the current top_parent and checking if mouse is in top parent	 
	@top_parent = top_parent if top_parent
	info_top = anchor_in_top_parent view, x, y
	@top_parent = nil if (info && !top_parent) || !info_top

	#Returning the info
	info
end

#ANCHOR: Retain the closest point between element and void picking
def anchor_compare_with_void(view, info, info_void)
	return info unless info_void
	return info_void unless info
	pt, type, info_comp = info
	ptvoid,	type_void = info_void
	return info if type_void == :free
		
	eye = view.camera.eye
	(eye.distance(pt) <= eye.distance(ptvoid)) ? info : info_void
end

#ANCHOR: Analyse the anchor and determine its type
def anchor_analysis(view, lst_anchor, x, y)
	type = nil
	nb_anchor = lst_anchor.length
	ptxy = Geom::Point3d.new x, y, 0
	
	i0, elt0, comp0, tr0 = lst_anchor.first
	i1, elt1, comp1, tr1 = lst_anchor[1]
	i2, elt2, comp2, tr2 = lst_anchor[2]
	
	#Center of Component or Group
	info = anchor_at_center_comp(view, x, y)
	return info if info
	
	#Guide Point
	if elt0.instance_of?(Sketchup::ConstructionPoint)
		info = [tr0 * elt0.position, :cpoint, [elt0, comp0, tr0]]
		
	#Guide Line
	elsif elt0.instance_of?(Sketchup::ConstructionLine)
		if elt1.instance_of?(Sketchup::ConstructionPoint)
			info = [tr1 * elt1.position, :cpoint, [elt1, comp1, tr1]]
		elsif nb_anchor == 1	
			info = anchor_on_guide_line(view, ptxy, elt0, comp0, tr0)
		elsif elt2.instance_of?(Sketchup::ConstructionPoint)
			info = [tr2 * elt2.position, :cpoint, [elt2, comp2, tr2]]
		else
			info = anchor_inter_cline(view, ptxy, elt0, comp0, tr0, elt1, comp1, tr1)		
		end	
		
	#Edge picked	
	elsif elt0.instance_of?(Sketchup::Edge)
		if nb_anchor == 1 || (comp0 == comp1 && elt1.instance_of?(Sketchup::Face) && elt1.edges.include?(elt0))
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		elsif comp0 == comp1 && elt1.instance_of?(Sketchup::Edge)
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		elsif elt1.instance_of?(Sketchup::Face) && !elt1.edges.include?(elt0)
			info = anchor_inter_edge_face(view, ptxy, elt0, comp0, tr0, elt1, comp1, tr1)
		else	
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		end
		
	#Face picked	
	elsif elt0.instance_of?(Sketchup::Face)
		if nb_anchor == 1
			info = anchor_on_face(view, ptxy, elt0, comp0, tr0)
		elsif nb_anchor == 2 && comp0 == comp1 && elt1.instance_of?(Sketchup::Edge) && elt0.edges.include?(elt1)
			info = anchor_on_edge(view, ptxy, elt1, comp1, tr1)
		elsif elt1.instance_of?(Sketchup::Edge) && !elt0.edges.include?(elt1)
			info = anchor_inter_edge_face(view, ptxy, elt1, comp1, tr1, elt0, comp0, tr0)
		elsif elt1.instance_of?(Sketchup::Text)
			info = anchor_on_text(view, ptxy, elt1, comp1, tr1)
		end	
		info = anchor_on_face(view, ptxy, elt0, comp0, tr0) unless info
	
	#Text picked	
	elsif elt0.instance_of?(Sketchup::Text)
		info = anchor_on_text(view, ptxy, elt0, comp0, tr0)

	#Dimension picked	
	elsif elt0.is_a?(Sketchup::Dimension)
		@ip.pick view, x, y
		pt = @ip.position.project_to_line view.pickray(x,y)
		info = [pt, :dimension, [elt0, comp0, tr0]]

	#Model Axes	and other elements
	elsif elt0.instance_of?(Sketchup::Drawingelement)
		if !comp0 && elt0.bounds.diagonal == 0
			info = anchor_on_axis(view, ptxy)
		else
			@ip.pick view, x, y
			pt = @ip.position.project_to_line view.pickray(x,y)
			info = [pt, :dimension, [elt0, comp0, tr0]]
		end	

	else
		@ip.pick view, x, y
		pt = @ip.position.project_to_line view.pickray(x,y)
		info = [pt, :element, [elt0, comp0, tr0]]
	end
	
	info
end

#ANCHOR: try to find a point
def anchor_in_void(view, x, y, info_def=nil)
	ptref, type_ref, info_comp_ref = @origin_info
	
	#Check if at center of a Grouponent
	info = anchor_at_center_comp(view, x, y)
	return info if info
	
	#Check if at center of a face
	info = anchor_at_center_face(view, x, y)
	return info if info
	
	#Check if at center of a arc
	info = anchor_at_center_arc(view, x, y)
	return info if info

	#Bounding box of groups and components
	info = anchor_at_bbox_comp(view, x, y)
	return info if info
	
	#No point of reference
	unless ptref || info_def
		view.tooltip = nil
		return nil
	end	
	
	#Forced direction
	return nil if @forced_vector_dir || @forced_plane_dir || @mode_origin
	
	#Return default direction if any
	return info_def if info_def
	
	#Other cases
	if ptref
		plane = [ptref, Z_AXIS]
		ray = view.pickray x, y
		target = Geom.intersect_line_plane ray, plane
		view.tooltip = nil
		return [target, :free, nil]	
	end
	nil	
end

#ANCHOR: Check if the mouse is within the bounding box of a component or group
def anchor_in_top_parent(view, x, y)
	@info_top_parent = nil
	return nil unless @top_parent
	bbox = G6.grouponent_definition(@top_parent).bounds
	tr = @top_parent.transformation
	
	#List of panels of the bounding box
	corners3d = G6.grouponent_corners(@top_parent, @top_parent.transformation)
	corners2d = corners3d.collect { |pt| p = view.screen_coords(pt) ; p.z = 0 ; p }
	ls_panels = []
	flat = (bbox.corner(0) == bbox.corner(4))
	if flat
		ls_panels.push [0, 1, 3, 2]
	else
		ls_panels.push [0, 1, 3, 2], [0, 1, 5, 4], [1, 3, 7, 5], [3, 2, 6, 7], [2, 0, 4, 6], [4, 5, 7, 6]
	end

	#Checking in 3D
	ray = view.pickray x, y
	eye = view.camera.eye
	
	ptxy = Geom::Point3d.new x, y, 0
	info_inter = nil
	dmin = nil
	ls_panels.each_with_index do |ipts, ipanel|
		pts = ipts.collect { |i| corners3d[i] }
		pts2d = ipts.collect { |i| corners2d[i] }
		plane = Geom.fit_plane_to_points pts[0], pts[1], pts[2]
		ptinter = Geom.intersect_line_plane ray, plane
		next unless ptinter
		d = eye.distance ptinter
		if (!dmin || d < dmin) && Geom.point_in_polygon_2D(ptxy, pts2d, true)
			info_inter = [ptinter, @top_parent, ipanel, pts, pts2d]
			dmin = d
		end
	end
	
	#No intersection found
	return nil unless info_inter
	@info_top_parent = info_inter

	#Otherwise on a face
	ptinter, ptxy, pts, pts2d = info_inter
	[ptinter, :bbox_face, [@top_parent, nil, tr]]
end

#ANCHOR: Test predefined directions
def anchor_follow_direction(view, x, y, ptref)
	ls_info = @origin_privileged_directions
	return unless ls_info && ptref
	
	#Calculate origin and target in 2D
	ptxy = Geom::Point3d.new x, y, 0
	origin2d = view.screen_coords ptref
	origin2d.z = 0
	vec2d = origin2d.vector_to ptxy
	size = view.pixels_to_model 30, ptref

	#Cycling through the possible direction inference
	ecart_min = @inference_distance
	result = nil
	ls_info.each do |symb, vec3d|
		pt = ptref.offset vec3d, size
		pt2d = view.screen_coords pt
		pt2d.z = 0
		axis2d = origin2d.vector_to pt2d
		axis2d = axis2d.reverse if axis2d % vec2d < 0
		next if vec2d.angle_between(axis2d) > @inference_angle
		d = origin2d.distance ptxy
		pt2d = origin2d.offset axis2d, d
		ecart = ptxy.distance(pt2d)
		if !ecart_min || ecart < ecart_min
			ecart_min = ecart
			result = [symb, vec3d]
		end	
	end
	
	#Calculating the target if there is a direction matching
	if result
		symb, vec3d = result
		lpt = Geom.closest_points [ptref, vec3d], view.pickray(x, y)
		target = lpt.first
		return [target, symb] if target	
	end	
	nil
end

#ANCHOR: Check if the target is at a center of a face
def anchor_at_center_face(view, x, y)
	return nil unless @center_face_info
	center, normal, info_face = @center_face_info
	center2d = view.screen_coords center
	if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
		return [center, :center_face, info_face]
	end
	nil
end

#ANCHOR: Check if the target is at a center of an arccurve
def anchor_at_center_arc(view, x, y)
	return nil unless @hsh_arc_center_info
	@hsh_arc_center_info.each do |id, info|
		center, curve, info_comp = info
		center2d = view.screen_coords center
		if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
			code = (curve.vertices.first == curve.vertices.last) ? :center_polygon : :center_arc
			return [center, code, info_comp]
		end
	end	
	nil
end

#ANCHOR: Check if the target is at a center of Component or Group
def anchor_at_center_comp(view, x, y)
	return nil unless @hsh_comp_center_info
	@hsh_comp_center_info.each do |id, info|
		center, comp, info_comp = info
		top_parent = info_comp[1]
		center2d = view.screen_coords center
		if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
			if comp.instance_of?(Sketchup::Group)
				code = (comp == top_parent) ? :center_group_top : :center_group
			else
				code = (comp == top_parent) ? :center_comp_top : :center_comp
			end	
			return [center, code, info_comp]
		end
	end	
	nil
end

#ANCHOR: Check if the target is at a corner of face center of the bounding box of Component or Group
def anchor_at_bbox_comp(view, x, y)
	return nil unless @hsh_comp_bbox_info
	ptxy = Geom::Point3d.new x, y, 0
	
	info_corner = nil
	dmin = @precision + 1
	@hsh_comp_bbox_info.each do |id, info|
		corners, comp, info_comp = info
		corners.each do |pt|
			pt2d = view.screen_coords pt
			pt2d.z = 0
			d = ptxy.distance pt2d
			if d < dmin
				dmin = d
				info_corner = [pt, :bbox_corner, info_comp]
			end	
		end
	end	
	info_corner
end

#ANCHOR: Check if the target is at Origin or on an axis
def anchor_on_axis(view, ptxy)
	#Test at ORIGIN
	origin2d = view.screen_coords ORIGIN
	origin2d.z = 0
	if (origin2d.x - ptxy.x).abs <= @precision  && (origin2d.y - ptxy.y).abs <= @precision
		return [ORIGIN, :origin ]
	end
	
	#Test on axes
	size = view.pixels_to_model 50, ORIGIN
	dmin = @precision
	info = nil
	[[:x, X_AXIS], [:y, Y_AXIS], [:z, Z_AXIS]].each do |symb, vec|
		pt2d = view.screen_coords ORIGIN.offset(vec, size)
		pt2d.z = 0
		d = ptxy.distance_to_line [origin2d, pt2d]
		if d <= dmin
			ptproj = ptxy.project_to_line [origin2d, pt2d]
			lpt = Geom.closest_points [ORIGIN, vec], view.pickray(ptproj.x, ptproj.y)
			if lpt
				info = [lpt.first, symb]
				dmin = d
			end	
		end			
	end
	(info) ? info : nil
end

#ANCHOR: Check if the target is on a GUIDE LINE
def anchor_on_guide_line(view, ptxy, cline, comp, tr)
	line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
	lpt = Geom.closest_points line, view.pickray(ptxy.x, ptxy.y)
	(lpt) ? [lpt.first, :on_cline, [cline, comp, tr]] : nil
end

#ANCHOR: Check if the target is on a GUIDE LINE
def anchor_on_text(view, ptxy, sutext, comp, tr)	
	#Close to the anchor
	pt_anchor = tr * sutext.point
	pt2d_anchor = view.screen_coords pt_anchor
	if pt2d_anchor.distance(ptxy) <= @precision
		return [pt_anchor, :text_anchor, [sutext, comp, tr]]
	end
	
	#On the Leader line
	pt_lab = pt_anchor.offset G6.transform_vector(sutext.vector, tr)
	pt2d_lab = view.screen_coords pt_lab
	d, = G6.proximity_point_segment(ptxy, pt2d_anchor, pt2d_lab)
	if d <= @precision
		lpt = Geom.closest_points [pt_anchor, pt_lab], view.pickray(ptxy.x, ptxy.y)
		return [lpt[0], :text_leader, [sutext, comp, tr]] if lpt&& lpt[0]
	end
	
	#On the Text label
	plane = [pt_lab, view.camera.direction]
	pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), [pt_lab, view.camera.direction]
	(pt) ? [pt, :on_text, [sutext, comp, tr]] : nil
end

#ANCHOR: Check if the target is on an EDGE
def anchor_on_edge(view, ptxy, edge, comp, tr)
	dmin = @precision
	info = nil
	
	#Checking vertices and midpoint of edge
	vx1 = edge.start
	vx2 = edge.end
	pt1 = tr * vx1.position
	pt2 = tr * vx2.position
	ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
	[[:vertex, pt1, vx1], [:vertex, pt2, vx2], [:midpoint, ptmid, edge]].each do |symb, pt, e|
		pt2d = view.screen_coords(pt)
		pt2d.z = 0
		d = ptxy.distance pt2d
		if d <= dmin
			info = [pt, symb, [e, comp, tr]]
			dmin = d
		end	
	end
	
	#Checking is close to edge line
	unless info
		lpt = Geom.closest_points [pt1, pt2], view.pickray(ptxy.x, ptxy.y)
		pt = lpt[0]
		info = [pt, :on_edge, [edge, comp, tr]] if pt
	end	

	info
end

#ANCHOR: Check if the target is on a FACE
def anchor_on_face(view, ptxy, face, comp, tr)
	#Test if close to center of face
	unless !@rendering_options["DrawHidden"] && face.edges.find { |e| e.smooth? || e.soft? || e.hidden? }
		info = anchor_at_center_face(view, ptxy.x, ptxy.y)
		return info if info
	end
	
	#Testing if a Comp center if on the face
	info = anchor_at_center_comp(view, ptxy.x, ptxy.y)
	return info if info

	#Testing if an arc center if on the face
	info = anchor_at_center_arc(view, ptxy.x, ptxy.y)
	return info if info

	#Position on the face
	pt1 = tr * face.vertices[0].position
	normal = G6.transform_vector(face.normal, tr)
	pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), [pt1, normal]
	
	(pt) ? [pt, :on_face, [face, comp, tr]] : nil
end

#ANCHOR: Check if the target is at the intersection of an Edge and a FACE
def anchor_inter_edge_face(view, ptxy, edge, compe, tre, face, compf, trf)
	normal = G6.transform_vector(face.normal, trf)
	plane = [trf * face.vertices[0].position, normal]
	pt1e = tre * edge.start.position
	pt2e = tre * edge.end.position
	veced = pt1e.vector_to pt2e
	
	#Edge parallel to face
	point, type = info_edge = anchor_on_edge(view, ptxy, edge, compe, tre)
	if pt1e.on_plane?(plane) && pt2e.on_plane?(plane)
		return info_edge
	elsif type && type != :on_edge
		pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), plane
		eye = view.camera.eye
		return info_edge if eye.distance(pt) >= eye.distance(point)
	end
	
	#Intersection Edge with Face
	pt = Geom.intersect_line_plane [pt1e, pt2e], [trf * face.vertices[0].position, normal]
	
	#Checking if real intersection
	if pt
		if [1, 2, 4].include?(face.classify_point(trf.inverse * pt))
			return [pt, :inter_edge_face, [edge, compe, tre], [face, compf, trf]]
		else
			return info_edge
		end
	end	
	
	#Point on face
	anchor_on_face(view, ptxy, face, compf, trf)
end

#ANCHOR: Check if the target is at the intersection of an Edge and a FACE
def anchor_inter_cline(view, ptxy, cline, comp, tr, elt, comp2, tr2)
	line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
	
	#Intersection Cline with Face
	if elt.instance_of?(Sketchup::Face)
		face = elt
		normal = G6.transform_vector(face.normal, tr)
		plane = [tr2 * face.vertices[0].position, normal]
		pt = Geom.intersect_line_plane line, plane
		if pt && [1, 2, 4].include?(face.classify_point(tr2.inverse * pt))
			return [pt, :inter_cline_face, [cline, comp, tr], [face, comp2, tr2]]
		end
	
	#Intersection Cline with Edge
	elsif elt.instance_of?(Sketchup::Edge)
		edge = elt
		eline = [tr2 * edge.start.position, tr2 * edge.end.position]
		pt = Geom.intersect_line_line line, eline
		if pt
			return [pt, :inter_cline_edge, [cline, comp, tr], [edge, comp2, tr2]]
		end

	#Intersection Cline with Cline
	elsif elt.instance_of?(Sketchup::ConstructionLine) && elt != cline
		pt = Geom.intersect_line_line line, [tr2 * elt.position, G6.transform_vector(elt.direction, tr2)]
		if pt
			return [pt, :inter_cline_cline, [cline, comp, tr], [elt, comp2, tr2]]
		end
	end
	
	#Point on cline (default)
	anchor_on_guide_line(view, ptxy, cline, comp, tr)
end

#ANCHOR: Compute the tooltip based on the anchor type
def anchor_tooltip(info)
	target, type, info_comp = info
	comp = (info_comp.class == Array) ? info_comp[1] : nil

	#Tip for grouponent
	if type == :center_comp_top || type == :center_group_top
		tip_comp = nil
	elsif comp.instance_of?(Sketchup::Group)
		tip_comp = T6[:INFER_InGroup]
	elsif comp.instance_of?(Sketchup::ComponentInstance)
		tip_comp = T6[:INFER_InComponent]
	elsif comp.instance_of?(Sketchup::Image)
		tip_comp = T6[:INFER_InImage]
	else	
		tip_comp = nil
	end

	#Tip for element
	tip = nil
	case type
	when :cpoint
		tip = T6[:INFER_Cpoint]
	when :on_cline
		tip = T6[:INFER_Cline]
	when :vertex
		tip = T6[:INFER_Endpoint]
	when :midpoint
		tip = T6[:INFER_Midpoint]
	when :center_face
		tip = T6[:INFER_CenterFace]
	when :center_arc
		tip = T6[:INFER_CenterArc]
	when :center_comp
		tip = T6[:INFER_CenterComp]
	when :center_comp_top
		tip = T6[:INFER_CenterCompTop]
	when :center_group
		tip = T6[:INFER_CenterGroup]
	when :center_group_top
		tip = T6[:INFER_CenterGroupTop]
	when :center_polygon
		tip = T6[:INFER_CenterPolygon]
	when :on_edge
		tip = T6[:INFER_OnEdge] 
	when :on_face
		tip = T6[:INFER_OnFace]
	when :origin
		tip = T6[:INFER_AtOrigin]
	when :x
		tip = T6[:INFER_OnXAxis]
	when :y
		tip = T6[:INFER_OnYAxis]
	when :z
		tip = T6[:INFER_OnZAxis]
	when :x, :follow_x
		tip = T6[:INFER_AlongXAxis]
	when :y, :follow_y
		tip = T6[:INFER_AlongYAxis]
	when :z, :follow_z
		tip = T6[:INFER_AlongZAxis]
	when :follow_z_comp
		tip = T6[:INFER_AlongZAxisComp]
	when :follow_x_comp
		tip = T6[:INFER_AlongXAxisComp]
	when :follow_y_comp
		tip = T6[:INFER_AlongYAxisComp]
	when :perp_edge
		tip = T6[:INFER_PerpEdgeFace]
	when :follow_edge
		tip = T6[:INFER_AlongEdgeDirection]
	when :plane_of_edges
		tip = T6[:INFER_PerpTwoEdges]
	when :follow_cline
		tip = T6[:INFER_AlongCline]
	when :follow_bbox
		tip = T6[:INFER_AlongAxesBbox]
	when :follow_normal
		tip = T6[:INFER_AlongNormalFace]
	when :inter_edge_face
		tip = T6[:INFER_InterEdgeFace]
	when :inter_cline_face
		tip = T6[:INFER_InterClineFace]
	when :inter_cline_edge
		tip = T6[:INFER_InterClineEdge]
	when :inter_cline_cline
		tip = T6[:INFER_InterClineCline]
	when :inter_vector_face
		tip = T6[:INFER_InterVectorFace]
	when :inter_plane_edge
		tip = T6[:INFER_InterPlaneEdge]
	when :text_anchor
		tip = T6[:INFER_TextAnchor]
	when :on_text
		tip = T6[:INFER_OnTextLabel]
	when :text_leader
		tip = T6[:INFER_OnTextLeader]
	when :dimension
		tip = T6[:INFER_OnDimension]
	when :element
		tip = T6[:INFER_OnDrawingElement]
	when :bbox_corner
		tip = T6[:INFER_AtCornerBbox]
	when :bbox_face
		tip = "on Face of Bounding Box"
	end
	
	#Full tip text
	if tip
		tip = tip + ((tip_comp) ? " [#{tip_comp}]" : "")
	end
	
	tip
end

#---------------------------------------------------------
# INFERENCE: Manage inferences
#---------------------------------------------------------

def inference_center_face(face, comp, tr)
	#@center_face_info = nil
	return unless face.valid?
	unless !@rendering_options["DrawHidden"] && face.edges.find { |e| e.smooth? || e.soft? || e.hidden? }
		@center_face_info = [tr * face.bounds.center, G6.transform_vector(face.normal, tr), [face, comp, tr]]
	end	
end

#INFERENCE: Store the center of arc curves for display and inference when specified by an edge
def inference_center_arc_from_edge(edge, comp, tr)
	curve = edge.curve
	return unless curve.instance_of?(Sketchup::ArcCurve)
	
	@hsh_arc_center_info = {} unless @hsh_arc_center_info
	
	id = "#{curve.entityID}-#{tr.to_a}"
	@hsh_arc_center_info[id] = [tr * curve.center, curve, [edge, comp, tr]]
end

#INFERENCE: Store the center of arc curves for display and inference when specified by a vertex
def inference_center_arc_from_vertex(vx, comp, tr)
	vx.edges.each { |edge| inference_center_arc_from_edge(edge, comp, tr) }
end

#INFERENCE: Store the center of component or group
def inference_center_grouponent(top_parent, comp, tr)
	@hsh_comp_center_info = {} unless @hsh_comp_center_info
	
	if top_parent && !top_parent.instance_of?(Sketchup::Image)
		id = top_parent.entityID
		@hsh_comp_center_info[id] = [top_parent.bounds.center, top_parent, [top_parent, top_parent, @tr_id]]
	end
	
	if comp && comp != top_parent && !comp.instance_of?(Sketchup::Image)
		id = comp.entityID
		t = tr * comp.transformation.inverse
		@hsh_comp_center_info[id] = [t * comp.bounds.center, comp, [comp, top_parent, t]]
	end
end

#INFERENCE: Store the corner of bounding boxes of component or group
def inference_bbox_grouponent(top_parent, comp, tr)
	@hsh_comp_bbox_info = {} unless @hsh_comp_bbox_info
	
	if top_parent && !top_parent.instance_of?(Sketchup::Image)
		id = top_parent.entityID
		corners = G6.grouponent_corners(top_parent, top_parent.transformation)
		@hsh_comp_bbox_info[id] = [corners, top_parent, [top_parent, top_parent, tr]]
	end
	
	if comp && comp != top_parent && !comp.instance_of?(Sketchup::Image)
		id = comp.entityID
		corners = G6.grouponent_corners(comp, tr)
		@hsh_comp_bbox_info[id] = [corners, comp, [comp, top_parent, tr]]
	end
end

#INFERENCE: Reset the position of the centers at their original position
def inference_update_centers
	#Face center
	if @center_face_info
		center, normal, face_info = @center_face_info
		inference_center_face(*face_info)
	end

	#Centers for Arc curves
	if @hsh_arc_center_info
		@hsh_arc_center_info.each do |id, info|
			center, edge, curve_info = info
			inference_center_arc_from_edge(*curve_info)
		end
	end

	#Centers for Components and Groups
	if @hsh_comp_center_info
		hsh = {}
		@hsh_comp_center_info.each do |id, info|
			center, comp, comp_info = info
			comp, parent, tr = comp_info
			id = "#{comp.entityID} - #{tr.to_a}"
			hsh[id] = [tr * comp.bounds.center, comp, comp_info]
		end
		@hsh_comp_center_info = hsh
	end
end

#INFERENCE: Remove all defined centers
def inference_reset_centers
	@center_face_info = @hsh_arc_center_info = @hsh_comp_center_info = @hsh_comp_bbox_info = nil
end

#---------------------------------------------------------
# DRAW: Manage drawing
#---------------------------------------------------------

#DRAW: Draw the origin, targets and joining line	
def draw_origin(view, flg_ip=true)
	draw_centers view
	draw_highlight_edge(view, @origin_info) unless @pt_target
	
	#Drawing the origin
	if flg_ip && @pt_origin && @ip_origin && @ip_origin.valid? && @ip_origin.display?
		@ip_origin.draw view 
	elsif @origin_info
		mark_draw view, @origin_info 
	end
end

#DRAW: Draw the origin, targets and joining line	
def draw_target(view, flg_ip=true)
	draw_centers view
	
	#Drawing the origin
	if flg_ip && @pt_target && @ip_target && @ip_target.valid? && @ip_target.display?
		@ip_target.draw view 
	elsif @pt_target
		target, symb_target, info_target = @target_info
		drive, symb_drive, info_drive = @info_drive3d
		draw_highlight_edge(view, @info_drive3d)
		if symb_target != symb_drive
			mark_draw view, [@pt_target, symb_target, info_target]
		end	
		mark_draw view, [@pt_drive3d, symb_drive, info_drive]
	end
end
	
#DRAW: Draw the origin, targets and joining line	
def draw_origin_target(view)
	return unless @pt_origin && @pt_target

	#Drawing the line between the origin and target
	vec = @pt_origin.vector_to @pt_target
	return unless vec.valid? 
	
	origin2d = view.screen_coords @pt_origin
	target2d = view.screen_coords @pt_target
	ptdrive2d = (@pt_drive3d) ? view.screen_coords(@pt_drive3d) : @pt_drive2d
	origin2d.z = target2d.z = ptdrive2d.z = 0
	if (@forced_vector_dir || @forced_plane_dir) && ptdrive2d != target2d
		code = (@forced_vector_dir) ? @forced_vector_code : @forced_plane_code
		color = @hsh_color_vector[code]
		view.line_width = 1
		view.line_stipple = '.'
		view.drawing_color = color
		view.draw2d GL_LINE_STRIP, [ptdrive2d, target2d]				
	end
	pt, symb = @info_drive3d
	if @forced_vector_dir
		view.line_width = 3
		code = @forced_vector_code
	elsif symb.to_s =~ /follow/
		view.line_width = 2
		code = symb
	else	
		view.line_width = 2
		code = (@target_info) ? @target_info[1] : 'free'
	end	

	view.line_stipple = '_'
	view.drawing_color = @hsh_color_vector[code]
	view.draw2d GL_LINE_STRIP, [origin2d, target2d]
end
	
#DRAW: Draw the distance
def draw_distance(view)
	return unless @pt_origin && @pt_target
	d = @pt_origin.distance @pt_target
	tip = G6.format_length_general(d) + ' '
	pt2d = @pt_drive2d
	bkcolor, frcolor = @color_box_distance
	G6.draw_rectangle_text(view, pt2d.x, pt2d.y, tip, bkcolor, frcolor)
end

#DRAW: Draw the Vector direction if any
def draw_vector_direction(view)
	return unless @origin_info && @forced_vector_dir
	ptori, code = @origin_info

	#Computing the guide line in 2D
	pt = ptori.offset @forced_vector_dir, 1
	pt1_2d = view.screen_coords ptori
	pt2_2d = view.screen_coords pt
	vec2d = pt1_2d.vector_to pt2_2d
	pt1 = pt1_2d.offset vec2d, view.vpwidth * 2
	pt2 = pt1_2d.offset vec2d, -view.vpwidth * 2
	
	#Drawing the guide line
	view.line_width = (@pt_target) ? 1 : 2
	view.line_stipple = '_'
	symb = @forced_vector_code
	view.drawing_color = @hsh_color_vector[symb]
	view.draw2d GL_LINE_STRIP, [pt1, pt2]
end

#DRAW: Draw the visual plane if applicable
def draw_plane_direction(view)
	draw_vector_direction(view)
	return unless @origin_info && @forced_plane_dir
	ptori, code = @origin_info
	
	#Size and axes of the plane
	axes = @forced_plane_dir.axes
	if @pt_target
		size = ptori.distance(@pt_target) * 1.2	
	else
		size = view.pixels_to_model(50, ptori)
	end
	size2 = size * 2
	
	#Computing the points
	vecx = axes[0]
	vecy = axes[1]
	pt1 = ptori.offset(vecx, size).offset(vecy, size)
	pt2 = pt1.offset vecx, -size2
	pt3 = pt2.offset vecy, -size2
	pt4 = pt3.offset vecx, size2
	pts = [pt1, pt2, pt3, pt4]
	
	#Drawing the plane
	symb = @forced_plane_code
	view.line_width = 2
	view.line_stipple = ''
	view.drawing_color = @hsh_bkcolor_plane[symb]
	view.draw GL_POLYGON, pts
	view.drawing_color = @hsh_frcolor_plane[symb]
	view.draw GL_LINE_LOOP, pts
end

#DRAW: Draw the curves when selection is a single vertex
def draw_hi_curve(view)
	return unless @hi_curve_pts
	view.line_width = 2
	view.line_stipple = ''
	view.drawing_color = 'orchid'
	@hi_curve_pts.each do |pts|
		view.draw GL_LINE_STRIP, pts.collect { |pt| G6.small_offset(view, pt) }
	end	
end

#DRAW: Draw marks on edges under target
def draw_highlight_edge(view, info)
	target, symb, info_comp = info
	return unless target 
	return unless symb == :on_edge || symb == :midpoint || symb == :vertex

	if symb == :vertex
		vx, comp, tr = info_comp
		lvx = []
		vx.edges.each do |edge|
			lvx.push edge.other_vertex(vx)
		end			
	else
		edge, comp, tr = info_comp
		lvx = [edge.start, edge.end]
	end
	
	quads = []
	eye = view.camera.eye
	lvx.each do |vx|
		pt = tr * vx.position
		size = view.pixels_to_model 3, pt
		size2 = 2 * size
		vecx, vecy = (eye.vector_to(pt)).axes
		pt0 = pt.offset vecx, size
		pt1 = pt0.offset vecy, size
		pt2 = pt1.offset vecx, -size2
		pt3 = pt2.offset vecy, -size2
		pt4 = pt3.offset vecx, size2
		quads += [pt1, pt2, pt3, pt4].collect { |pt| G6.small_offset view, pt }
	end
	
	color = 'lightblue'
	view.drawing_color = color
	view.draw GL_QUADS, quads
end

#DRAW: Draw the center of the active face and arc curves if applicable
def draw_centers(view)
	#Draw the top_axes
	draw_top_axes view

	#Draw the bounding boxes for components and groups
	if @hsh_comp_bbox_info
		@hsh_comp_bbox_info.each do |id, info|
			corners, comp, info_comp = info
			next unless comp.valid?
			comp, top_parent, tr = info_comp
			lix = (corners.length == 4) ? @bbox_lines_indexing[0..7] : @bbox_lines_indexing
			lines = lix.collect { |i| corners[i] }
			view.line_width = 1
			view.line_stipple = ''
			view.drawing_color = 'gray'
			view.draw GL_LINES, lines				
		end
	end		
		
	#Draw the center of face
	center_face, normal = @center_face_info
	if center_face
		size = view.pixels_to_model 6, center_face
		vecx, vecy, = normal.axes
		pts = [center_face.offset(vecx, -size), center_face.offset(vecx, size), center_face.offset(vecy, -size), center_face.offset(vecy, size)]
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = 'blue'
		view.draw2d GL_LINES, pts.collect { |pt| view.screen_coords pt }	
	end
	
	#Draw the center of arc curves
	if @hsh_arc_center_info
		@hsh_arc_center_info.each do |id, info|
			center, curve, info_comp = info
			next if center == center_face
			return unless center
			mark_draw view, [center, :center_arc_mark, info_comp]
		end	
	end
	
	#Draw the centers for Component or groups
	if @hsh_comp_center_info
		@hsh_comp_center_info.each do |id, info|
			center, comp, info_comp = info
			top_parent = info_comp[1]
			next if center == center_face
			return unless center
			code = (top_parent == comp) ? :center_comp_mark_top : :center_comp_mark
			mark_draw view, [center, code, info_comp]
		end	
	end		
end

#DRAW: draw the top axes for component
def draw_top_axes(view)
	return unless @top_axes && @top_parent
	tr = @top_parent.transformation
	center = tr * @top_parent.definition.insertion_point
	center2d = view.screen_coords center
	center2d.z = 0
	size = view.pixels_to_model 50, center
	view.line_stipple = ''
	view.line_width = 3
	[[X_AXIS, 'red'], [Y_AXIS, 'green'], [Z_AXIS, 'blue']].each do |vec, color|
		pt = center.offset G6.transform_vector(vec, tr), size
		pt2d = view.screen_coords pt
		pt2d.z = 0
		view.drawing_color = color
		view.draw2d GL_LINE_STRIP, [center2d, pt2d]
	end
end

#---------------------------------------------------------
# SELECTION: Manage selection and auto-selection
#---------------------------------------------------------

#SELECTION: Auto selection where there is no pre-selection	
def auto_selection(view, x, y, extended, multi_selection, no_vertex)
	return @preselection if @preselection
	
	#There is a top parent identified
	if @top_parent
		suselection_clear unless multi_selection || @cumulative_selection
		suselection_add @top_parent
		@active_selection = @selection.to_a
		preselection_cumulate if multi_selection
		return @active_selection
	end
	
	#Picking the elements
	@ph.do_pick x, y, @precision
	
	hsh_elt = {}
	lst_element = []
	edge = face = elt = cpoint = nil
	@ph.all_picked.each do |e|
		if e.instance_of?(Sketchup::Edge)
			unless hsh_elt[:edge]
				lst_element.push e
				hsh_elt[:edge] = e
				edge = e
			end	
		elsif e.instance_of?(Sketchup::Face)
			unless hsh_elt[:face]
				lst_element.push e
				hsh_elt[:face] = e
				face = e
			end	
		elsif e.instance_of?(Sketchup::ConstructionPoint)
			unless hsh_elt[:cpoint]
				lst_element.push e
				hsh_elt[:cpoint] = e
				cpoint = e
			end	
		elsif (e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance))
			unless hsh_elt[:comp]
				lst_element.push e
				hsh_elt[:comp] = e
			end	
		elsif !hsh_elt[:elt]
			if !(e.instance_of?(Sketchup::Drawingelement) && e.bounds.diagonal == 0)	#discard model axes
				lst_element.push e
				hsh_elt[:elt] = e
				elt = e
			end			
		end	
	end
	
	#Retaining the good element picked
	sel = lst_element.first
	
	#Special case when edge or face are at the surface of a component
	[cpoint, edge, face, elt].each do |e|
		if e && e.parent == Sketchup.active_model
			sel = e
			break
		end	
	end	
	
	#Vertex only
	@hi_curve_pts = nil
	if sel.instance_of?(Sketchup::Edge) && !no_vertex && !extended && !@cumulative_selection && !multi_selection
		edge = sel
		@ip.pick view, x, y
		vx = @ip.vertex
		if vx
			curve = edge.curve
			unless curve && vx.curve_interior?
				lpts = []
				vx.edges.each do |ee|
					lvx = (ee.curve) ? ee.curve.vertices : [ee.start, ee.end]
					lpts.push.push lvx.collect { |v| v.position }
				end	
				@hi_curve_pts = lpts
				suselection_clear
				@active_selection = [vx]
				return @active_selection
			end	
		end	
	end	
	
	#Extending edges to curve
	if sel.instance_of?(Sketchup::Edge)
		if sel.curve
			sel = sel.curve.edges
		elsif extended
			anglemax = 40.degrees
			sel = [edge] + G6.curl_follow_extend(sel, anglemax, true)
		else
			sel = [sel]
		end	
	elsif sel.instance_of?(Sketchup::Face)
		sel = (@rendering_options["DrawHidden"] && !extended) ? [sel] : G6.face_neighbours(sel)
	elsif sel
		sel = [sel]
	end	
		
	#Finding centers of faces
	unless sel
		info = anchor_at_center_face(view, x, y)
		if info
			center, symb, info_face = info
			face, comp, tr = info_face
			if comp
				sel = [comp]
			elsif face
				sel = [face]
			end	
		end
	end

	#Finding centers of Component or Group
	unless sel
		info = anchor_at_center_comp(view, x, y)
		info = anchor_at_bbox_comp(view, x, y) unless info
		if info
			center, symb, info_comp = info
			comp, parent, tr = info_comp
			sel = (parent) ? [parent] : [comp]
		end
	end

	#Finding centers of arcs
	unless sel
		info = anchor_at_center_arc(view, x, y)
		if info
			center, symb, info_arc = info
			edge, comp, tr = info_arc
			if comp
				sel = [comp]
			elsif edge && edge.curve
				sel = edge.curve.edges
			end	
		end
	end
	
	#Replacement of additive mode depending on Shift	
	suselection_clear if !multi_selection
	suselection_add @cumulative_selection if @cumulative_selection
	
	#Updating the selection	
	if sel
		ledges = []
		sel.each do |face|
			face_edges_visible(face).each { |e| ledges.push e unless sel.include?(e) } if face.instance_of?(Sketchup::Face)
		end
		suselection_add sel + ledges
		@active_selection = @selection.to_a
		@active_selection = nil if @active_selection.empty?
	end
	
	@active_selection
end

#SUSELECTION: find the edges of a face when visible
def face_edges_visible(face)
	ledges = face.edges
	return ledges if @rendering_options["DrawHidden"]
	ledges.find_all { |e| e.faces.length == 1 || G6.edge_plain?(e) }
end
	
#SUSELECTION: Clear the Sketchup Selection
def suselection_clear
	return if @ignore_selection
	@selection.clear
end

#SUSELECTION: Manage the Sketchup Selection
def suselection_add(lentities, clear_before=false)
	return if @ignore_selection
	@selection.clear if clear_before
	return unless lentities
	lentities = [lentities] unless lentities.class == Array
	return unless lentities && lentities.length > 0
	
	#Adding the top-model elements to the selection
	@selection.add lentities
	
	#Computing the box inferences for groups and components
	ls_grouponents = lentities.grep(Sketchup::Group) + lentities.grep(Sketchup::ComponentInstance)	
	ls_grouponents.each do |comp|
		inference_bbox_grouponent(comp, nil, @tr_id)
	end

	@selection
end
	
end	#class OriginTargetPicker

end	#End Module Traductor
